# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.

from __future__ import annotations

import logging

__title__ = "tcsa"
__description__ = "temporalis segmentation pipeline to assess CSA of temporalis"
__url__ = "https://gitlab.com/computational.oncology/temporalis-segmentation-pipeline"
__author__ = """Computational Oncology"""
# __email__ = ""
__version__ = "0.1.1"
__license__ = "GPL 3"
__copyright__ = "Copyright 2022 Imperial College London"


logging.getLogger(__name__).addHandler(logging.NullHandler())

import requests
from packaging.version import parse, Version
import json

def get_version(url_pattern: str = 'https://pypi.org/pypi/tcsa/json') -> Version:
    """Returns version of tCSA on test.pypi.org using json.

    Args:
        url_pattern (str): url to the json of the package

    return:
        Version: latest version of the package
    """
    req = requests.get(url_pattern)
    version = parse('0')

    if req.encoding:
        if req.status_code == requests.codes.ok:
            j = json.loads(req.text.encode(req.encoding))
            releases = j.get('releases', [])

            for release in releases:
                ver = parse(release)
                if not ver.is_prerelease:
                    version = max(version, ver)

    return version

def welcome() -> None:
    """
    Print a welcome screen and check if the package need an update.
    """
    print(rf"""

                         Cross                             Area
______________________/\\\\\\\\\_____/\\\\\\\\\\\_______/\\\\\\\\\____
 ___________________/\\\////////____/\\\/////////\\\___/\\\\\\\\\\\\\__
  _____/\\\________/\\\/____________\//\\\______\///___/\\\/////////\\\_
   __/\\\\\\\\\\\__/\\\_______________\////\\\_________\/\\\_______\/\\\_
    _\////\\\////__\/\\\__________________\////\\\______\/\\\\\\\\\\\\\\\_
     ____\/\\\______\//\\\____________________\////\\\___\/\\\/////////\\\_
      ____\/\\\_/\\___\///\\\___________/\\\______\//\\\__\/\\\_______\/\\\_
       ____\//\\\\\______\////\\\\\\\\\_\///\\\\\\\\\\\/___\/\\\_______\/\\\_
        _____\/////__________\/////////____\///////////_____\///________\///__
           temporalis                        Sectional


                                    v{__version__}
                                License: GPL v3

                     *************************************
             Oncology Lab | Imperial College London | NHS Trust
                     *************************************

    """)
    latest = get_version('https://test.pypi.org/pypi/tcsa/json')
    if parse(__version__) < latest:
        print(rf"""
A new version of tCSA is available (v{str(latest)}).
You can update it by running:
$ pip install --upgrade tcsa
        """)
    else:
        print("You are using the latest version of tCSA.")


from pathlib import Path

##########################################################
##  Different paths necessary for the program to work:  ##
##  main:                                               ##
##  ├─── images: (your input folder)                    ##
##  │	 ├─── patient_name_01.nifti                     ##
##  │	 └─── ...                                       ##
##  ├─── slices: (folder for the slices)                ##
##  │    ├─── patient_name_01:                          ##
##  │    │    ├─── original:                            ##
##  │    │    │    ├─── nbslice.nifti                   ##
##  │    │    │    └─── ...                             ##
##  │    │    ├─── normalized:                          ##
##  │    │    │    ├─── nbslice.nifti                   ##
##  │    │    │    └─── ...                             ##
##  │    │    └─── preprocessed:                        ##
##  │    │    	   ├─── nbslice.nifti                   ##
##  │    │         └─── ...                             ##
##  │    └─── patient_name_...:                         ##
##  ├─── eyeballs:                                      ##
##  │    ├─── patient_name_01:                          ##
##  │    │    ├─── original:                            ##
##  │    │    │    ├─── nbslice.png                     ##
##  │    │    │    └─── ...                             ##
##  │    │    └─── segmentation:                        ##
##  │    │         ├─── nbslice.png                     ##
##  │    │         └─── ...                             ##
##  │    └─── patient_name_...:                         ##
##  ├─── temporalis:                                    ##
##  │    ├─── patient_name_01:                          ##
##  │    │    ├─── original:                            ##
##  │    │    │    ├─── nbslice.png                     ##
##  │    │    │    └─── ...                             ##
##  │    │    └─── segmentation:                        ##
##  │    │         ├─── nbslice.png                     ##
##  │    │         └─── ...                             ##
##  │    └─── patient_name_...:                         ##
##  └─── trained_weights:                               ##
##  	 ├─── unet_eyeball.hdf5                         ##
##  	 └─── unet_temporalis.hdf5                      ##
##########################################################

PATHS = {'main' : Path.cwd()}
# PATHS['images'] = PATHS['main'] / 'images'
PATHS['slices'] = PATHS['main'] / 'slices'
PATHS['eyeballs'] = PATHS['main'] / 'eyeballs'
PATHS['temporalis'] = PATHS['main'] / 'temporalis'
PATHS['weights'] = PATHS['main'] / 'trained_weights'



#!/usr/bin/env python
#
# Copyright (c) 2019 zhixuhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.

from pathlib import Path

# remove deprecated type warning
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

# remove warnings from tensorflow about deprecated features and only display errors
import logging
import tensorflow as tf
tf.get_logger().setLevel(logging.ERROR)

# remove warnings stating that the model is not optimized for cpu usage
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import sys
# remove 'Using TensorFlow backend'
stderr = sys.stderr
sys.stderr = open(os.devnull, 'w')

from keras.models import Model
sys.stderr = stderr

from keras.layers import (Conv2D, BatchNormalization, MaxPooling2D,
                          Dropout, UpSampling2D, concatenate)
from keras.engine import Input
from keras.optimizers import Adam

# remove 'Segmentation Models: using `keras` framework.'
stdout = sys.stdout
sys.stdout = open(os.devnull, 'w')
import segmentation_models as sm
sys.stdout = stdout

dice_loss = sm.losses.DiceLoss()

dice_coeff = sm.metrics.FScore(threshold=0.5)

def unet(pretrained_weights: Path = None, input_size: tuple = (256,256,1)) -> Model:
    # added batch normalization layers after each convolutional layer
    inputs = Input(input_size)
    conv1 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(inputs)
    conv1 = BatchNormalization()(conv1)
    conv1 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv1)
    conv1 = BatchNormalization()(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    conv2 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool1)
    conv2 = BatchNormalization()(conv2)
    conv2 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv2)
    conv2 = BatchNormalization()(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    conv3 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool2)
    conv3 = BatchNormalization()(conv3)
    conv3 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv3)
    conv3 = BatchNormalization()(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    conv4 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool3)
    conv4 = BatchNormalization()(conv4)
    conv4 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv4)
    conv4 = BatchNormalization()(conv4)
    drop4 = Dropout(0.5)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(drop4)

    conv5 = Conv2D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool4)
    conv5 = BatchNormalization()(conv5)
    conv5 = Conv2D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv5)
    conv5 = BatchNormalization()(conv5)
    drop5 = Dropout(0.5)(conv5)

    up6 = Conv2D(512, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(drop5))
    merge6 = concatenate([drop4,up6], axis = 3)
    conv6 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge6)
    conv6 = BatchNormalization()(conv6)
    conv6 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv6)
    conv6 = BatchNormalization()(conv6)

    up7 = Conv2D(256, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv6))
    merge7 = concatenate([conv3,up7], axis = 3)
    conv7 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge7)
    conv7 = BatchNormalization()(conv7)
    conv7 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv7)
    conv7 = BatchNormalization()(conv7)

    up8 = Conv2D(128, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv7))
    merge8 = concatenate([conv2,up8], axis = 3)
    conv8 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge8)
    conv8 = BatchNormalization()(conv8)
    conv8 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv8)
    conv8 = BatchNormalization()(conv8)

    up9 = Conv2D(64, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv8))
    merge9 = concatenate([conv1,up9], axis = 3)
    conv9 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge9)
    conv9 = BatchNormalization()(conv9)
    conv9 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv9)
    conv9 = BatchNormalization()(conv9)
    conv9 = Conv2D(2, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv9)
    conv9 = BatchNormalization()(conv9)
    conv10 = Conv2D(1, 1, activation = 'sigmoid')(conv9)

    model = Model(inputs = inputs, outputs = conv10)

    model.compile(optimizer = Adam(lr = 0.0005), loss = dice_loss, metrics = ['accuracy', dice_coeff])

    # model.summary()

    if(pretrained_weights):
        model.load_weights(pretrained_weights)

    return model



# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.

import argparse
from pathlib import Path
from hashlib import md5
import gdown
import json
import logging
import importlib.resources
import dicom2nifti
import sys
import shutil

from tcsa import PATHS
from .data_preparation import is_nifti, is_dicom

FORMAT = "%(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
log = logging.getLogger(__name__)

log_conv = logging.getLogger('dicom2nifti')
log_conv.setLevel(logging.ERROR)

def create_parser() -> argparse.ArgumentParser:
    """
    Create the parser used in CLI.

    Return:
        (argparse.ArgumentParser): Parser with desired argument.
    """
    parser = argparse.ArgumentParser(
        prog='tcsa',
        description='Extract the CSA of temporalis muscle from T1w-contrast MRI',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog='only for POSIX system for now.'
    )
    parser.add_argument(
        "images",
        metavar="path_to_image(s)",
        nargs="?",
        type=Path,
        help="Path of image(s) to process, can be a folder or single image (.nii or .nii.gz)."
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="path_to_csv",
        nargs="?",
        type=Path,
        default=(Path.cwd() / 'temporalis_CSA.csv'),
        help="Path to save the csv file."
    )
    parser.add_argument(
        "-d",
        "--delete",
        action="store_true",
        help="Only keep input and output."
    )
    parser.add_argument(
        "--both-sides",
        action="store_true",
        help="segment both temporalis muscle."
    )
    # parser.add_argument(
    #     "-v",
    #     "--verbosity",
    #     action="count",
    #     default=0,
    #     help="Increase output verbosity (e.g., -vv is more than -v)."
    # )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Display version of tCSA."
    )

    parser.add_argument(
        "--dicom",
        action="store_true",
        help="Convert dicom directories to nifti."
    )
    return parser

def handle_dicom(args: argparse.Namespace, *, verbosity: bool = False) -> None:
    """
    Convert dicom directory in nifti for the pipeline.

    Pipeline work for nifti files,
    so we convert dicom into nifti files beforehand,
    the new nifti file is stored in a temp directory

    Args:
        args (argparse.Namespace): Arguments from CLI.
        verbosity (bool): Control the verbosity of dicom2nifti
    """
    if args.images.is_dir():
        dicom_files = list(args.images.iterdir())
        if len(dicom_files) == 0:
            log.warning("There is no files in your directory.")
            sys.exit()
        if is_dicom(dicom_files[0]):
            if verbosity:
                log_conv.setLevel(logging.INFO)
            log.info("Convering dicom to nifti...")
            temp_images = Path('temp_images').resolve()
            temp_images.mkdir(parents=True, exist_ok=True)
            dicom2nifti.convert_directory(args.images, temp_images,
                                          compression=False, reorient=True)
            args.images = temp_images
        else:
            log.warning("No dicom files found in the directory.")
            sys.exit()
    else:
        log.warning("Please put the dicom files in a directory.")
        sys.exit()

def handle_single_nifti(args: argparse.Namespace) -> None:
    """
    Make the pipeline work for single nifti file.

    Pipeline work for directories, so if there is only one image
    we create a temporary directory where we copy the image
    the temp directory will be our new input

    Args:
        args (argparse.Namespace): Arguments from CLI.
    """
    if is_nifti(args.images):
        temp_images = Path('temp_images').resolve()
        temp_images.mkdir(parents=True, exist_ok=True)
        shutil.copy(str(args.images), str(temp_images / args.images.name))
        args.images = temp_images
    else:
        log.warning(f"{args.images.name} isn't a nifti file.")
        if is_dicom(args.images):
            log.warning(f"Please convert with --dicom in command line.")
        sys.exit()

def creating_directories() -> None:
    """
    Create the directories present in tcsa (__init__.py)
    """
    log.info("\nCreating the necessary directories...")
    for path in PATHS.values():
        path.mkdir(parents=True, exist_ok=True)

def get_hash(file: Path) -> str:
    """
    Get md5Hash of a file

    Args:
        file (Path): Path to file

    Return:
        str: md5Hash of file
    """
    md5_hash = md5()
    with open(str(file), 'rb') as f:
        for byte_block in iter(lambda: f.read(4096),b''):
            md5_hash.update(byte_block)
    return md5_hash.hexdigest()

def get_weights() -> None:
    """
    Check the weights and download them if necessary

    Raise:
        Exception: the weight downloaded is corrupted
    """
    # load configs containing the hash of the weight and there download link
    with importlib.resources.open_text("tcsa", "config.json") as config_file:
        config = json.load(config_file)
    HASHS, ID = config['HASHS'], config['ID']

    log.info("\nChecking if weights are present...")
    for weight, hash in HASHS.items():
        weight_path = PATHS['weights'] / weight
        if weight_path.exists():
            if get_hash(weight_path) == hash:
                log.info(f"{weight} is present and is up to date")
                continue
            else:
                log.info(f"{weight} is present but is corrupted/not up to date")
        else:
            log.info(f"{weight} is missing")
        log.info(f"{weight} will be downloaded")
        gdown.download(id=ID[weight], output=str(weight_path), quiet=False,
                       use_cookies=False, resume=False)
        if get_hash(weight_path) != hash:
            raise Exception("md5 checksum failed")

def run_prequisite() -> None:
    """
    Run all the prequisites (create directories and check weights).
    """
    creating_directories()
    get_weights()



# Copyright (c) 2019 zhixuhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.



from pathlib import Path
import numpy as np
import skimage.io as io
import skimage.transform as trans
from skimage import img_as_ubyte


def testGenerator(full_path: Path, target_size: tuple = (256,256)) -> np.ndarray:
    """
    Generator to feed the images to the neural network.

    Args:
        full_path (Path): path to eyeballs/patient/original,
                          where preprocessed png are stored
        target_size (tuple): size of the final image

    Return:
        img (np.ndarray): numpy array of the png image with the desired property
    """
    for filename in full_path.iterdir():
        img = io.imread(str(filename), as_gray = True)
        img = img / 255.
        img = trans.resize(img,target_size)
        img = np.reshape(img,img.shape+(1,))
        img = np.reshape(img,(1,)+img.shape)
        yield img


def saveResult(output_path: Path, npyfile: np.ndarray, list_name: list) -> None:
    """
    Save the results of the model prediction as png in output_path.

    Args:
        output_path (Path): Path to where the png segmentation will be stored
        npyfile (np.ndarray): numpy array of the model prediction
        list_name (list): list containing strings of the desired output filename
    """
    for name in list_name:
        output_path.mkdir(parents=True, exist_ok=True)
        for i, item in enumerate(npyfile):
            if list_name.index(name) == i:
                img = item[:,:,0]
                io.imsave(str(output_path / name), img_as_ubyte(img),
                                                           check_contrast=False)



# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.

from pathlib import Path
import os
from shutil import rmtree
from tqdm import tqdm
import sys

import numpy as np
import nibabel as nib
import ants
import imageio

from tcsa import PATHS

# suppress warning about losy conversion for imageio
import logging
logger = logging.getLogger('imageio')
logger.setLevel(logging.ERROR)

FORMAT = "%(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
log = logging.getLogger(__name__)

def is_nifti(path_to_image: Path) -> bool:
    """
    Check if the file is a nifti image.

    Take into account compressed nifti files

    Args:
        path_to_image (Path): Path to the file.
    Return:
        bool: return true if the file is a nifti file
    """
    extensions = ''.join(path_to_image.suffixes)
    return True if extensions in ('.nii', '.nii.gz') else False

def is_dicom(path_to_image: Path) -> bool:
    """
    Check if the file is a dicom image.

    Args:
        path_to_image (Path): Path to the file.
    Return:
        bool: return true if the file is a dicom file
    """
    extensions = ''.join(path_to_image.suffixes)
    return True if extensions in ('.dcm', '.DCM') else False

def valid_format(path_to_image: Path) -> bool:
    """
    Handle the different file format.

    Check if it's a nifti file, a dicom file,
    and provide information on what to do.

    Args:
        path_to_image (Path): Path to the file.
    Return:
        bool: return true if the file is a nifti file
    """
    if is_nifti(path_to_image):
        return True
    else:
        if is_dicom(path_to_image):
            log.warning(f'{path_to_image.name}: Please convert with --dicom in command line.')
        else:
            log.warning(f'{path_to_image.name}: Unknown file format.')
        return False

def slicing(path_to_images: Path,
            path_to_slices_folder: Path = PATHS['slices'],
            *, window: list = [0.25, 0.5]) -> None:
    """
    Slice the nifti file in path_to_images.

    Slice the images present in 'path_to_images',
    and output them in 'path_to_slices_folder' and are stored the following way:
    path_to_slices_folder:
    └─── original:
        ├─── patient_name:
        │   ├─── nbslice.extensions (nii or nii.gz)
        │   └─── ....
        └─── patient_name:
            ├─── nbslice.extensions
            └─── ....

    Args:
        path_to_images (Path): Path to the images (only Nifti images will work)
        path_to_slices_folder (Path): Ouput path of the slices
        window (list): only take into account a certain percentage of slices
                        (help reduce false eyeball segmentation)
    """
    if path_to_images.name == 'temp_images':
        log.info("\nSlicing image into 2D slices...")
    else:
        log.info("\nSlicing images into 2D slices...")
    for patient in path_to_images.iterdir():
        if not valid_format(patient):
            continue
        extensions = ''.join(patient.suffixes)
        patient_name = patient.name[: -len(extensions)]

        patient_directory = path_to_slices_folder / patient_name / 'original'
        patient_directory.mkdir(parents=True, exist_ok=True)

        img = nib.load(patient)
        nb_slices = img.shape[2]
        img = img.get_fdata()
        if len(img.shape) == 4:
            log.warning(f"{patient.name} contain multiple components (4D, or dicom). Only the first one will be kept")
            img = img[:,:,:,0]
        # img_cropped = img[..., nb_slices // 4:nb_slices // 2]
        for i in tqdm(
                range(round(nb_slices*window[0]), round(nb_slices*window[1])),
                desc=f"Slicing {patient.name}"
        ):
            imageio.imsave(str(patient_directory / f'{i}{extensions}'),
                                                           np.rot90(img[:,:,i]))
            # nib.save(img.slicer[..., i:i+1],
            #         path_to_slices_folder / f'{filename}_{i}{extensions}')



def preprocessing(path_to_slices_folder: Path = PATHS['slices'],
                  *, verbose: bool = False,
                  delete_og_slices: bool = False) -> None:
    """
    Preprocess nifti slices.

    Perform z-score normalization, bias field correction, resample, set spacing
    on slices and proceed to save them.

    Args:
        path_to_slices_folder (Path): Path to the root directory of the slices,
                                     contening ALL the slices (refer to README)
    Kwargs:
        verbose (bool): Make z-score verbose or not
        delete (bool): Delete unused slices (original and normalized)
    """
    nb_patients = len(list(path_to_slices_folder.iterdir()))
    if nb_patients == 0:
        log.warning("No slice found.")
        sys.exit()
    log.info("\nNormalizing slices ...")
    for patient in path_to_slices_folder.iterdir():
        slices_path = patient / 'original'
        normalized_path = patient / 'normalized'
        normalized_path.mkdir(parents=True, exist_ok=True)
        for slice in tqdm(
            slices_path.iterdir(),
            desc=f"Normalizing slices for {patient.name}",
            total=len(list(slices_path.iterdir()))
        ):
            normalization_query = (f"zscore-normalize '{str(slice)}'"
                                f" -o '{str(normalized_path / slice.name)}'")
            if verbose: normalization_query += ' -v'
            os.system(normalization_query)

    log.info("\nPreprocessing slices ...")
    for patient in path_to_slices_folder.iterdir():
        slices_path = patient / 'original'
        normalized_path = patient / 'normalized'
        preprocessed_path = patient / 'preprocessed'
        preprocessed_path.mkdir(parents=True, exist_ok=True)
        for slice in tqdm(
                        normalized_path.iterdir(),
                        desc=f"Preprocessing slices for {patient.name}",
                        total=len(list(normalized_path.iterdir()))
        ):
            img = nib.load(slice).get_fdata() # add [:,:,0] if the slice is still considered 3D, 2 main reasons possible:
                                              # you didn't use provided slicing method, or you used 4D images (3D and time)
            img = ants.from_numpy(img, origin = (0,0))
            img_n4 = ants.n4_bias_field_correction(img)
            img_res = ants.resample_image(img_n4, (256,256), True, 1)
            img_res.set_spacing((1,1))

            extensions = ''.join(slice.suffixes)
            filename = slice.name[: -len(extensions)]
            output_filename = f'{filename}{extensions}'
            output_path = preprocessed_path / output_filename
            ants.image_write(img_res, str(output_path))

        if delete_og_slices:
            rmtree(str(slices_path), ignore_errors=True)
            rmtree(str(normalized_path), ignore_errors=True)


def converting_to_png(path_to_slices: Path = PATHS['slices'],
                      path_to_test_folder: Path = PATHS['eyeballs']) -> None:
    """
    Convert nifti images to png to be fed to the neural network.

    Args:
        path_to_slices (Path): path that contains the nifti slices to convert
        path_to_test_foler (Path): ouput path that contains png of slices
    """
    log.info("\nConverting images to PNG....")
    for patient in path_to_slices.iterdir():
        preprocessed_path = patient / 'preprocessed'
        for slice in tqdm(
                        preprocessed_path.iterdir(),
                        desc=f"Converting to PNG for {patient.name}",
                        total=len(list(preprocessed_path.iterdir()))
        ):
            img = nib.load(slice).get_fdata()
            img = np.rot90(img[:,:], 3)
            filename = slice.name[: -len(''.join(slice.suffixes))]
            output_directory = path_to_test_folder / patient.name / 'original'
            output_directory.mkdir(parents=True, exist_ok=True)
            output_path = output_directory / f'{filename}.png'
            imageio.imsave(str(output_path), img)



# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.

import numpy as np
from pathlib import Path
import cv2
import shutil
import pandas as pd

from tcsa import PATHS

import logging
FORMAT = "%(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
log = logging.getLogger(__name__)


def select_slice(input_path: Path, output_path: Path,
                        *, nifti: bool = False,
                        delete_eyeballs: bool = False) -> None:
    """
    Select the slice at the right eyeball level

    First convert the segmentation.png present in input_path / original
    in a binary image and count the number of white pixel
    (corresponding to the area of the segmented eyeballs).
    Then calculate the 97th percentile of all the area for one patient and take
    the first slice greater than the 97th percentile and output the original png
    in output_path / original
    Possible to also get the nifti slice.

    Args:
        input_path (Path): Path to the png slice of the patient
        output_path (Path): Path for the png slice of the patient at the right level
    Kwargs:
        nifti (bool): if set to true also output the nifti slice
        delete (bool): delete all slices (eyeball png and nifti slices)
    """
    areas_list = []
    ordered_slices = sorted(input_path.iterdir(), key=lambda x: int(x.stem))
    for segmentation in ordered_slices:
        img = cv2.imread(str(segmentation))
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        _, thresh = cv2.threshold(gray, 127, 255, cv2.THRESH_BINARY)
        area = cv2.countNonZero(thresh)
        areas_list.append(area)
    area_97 = np.percentile(areas_list, 97)
    for i, area in enumerate(areas_list):
        if area >= area_97:
            eyeball_pictures = input_path.parent / 'original'
            eyeball = sorted(eyeball_pictures.iterdir(), key=lambda x: int(x.stem))[i]
            output_path /= 'original'
            output_path.mkdir(parents=True, exist_ok=True)
            shutil.copy(str(eyeball), str(output_path / eyeball.name))
            if nifti:
                patient_name = input_path.parent.name
                nifti_slice = PATHS['slices'] / patient_name / 'preprocessed' / f'{eyeball.stem}.nii.gz'
                shutil.copy(str(nifti_slice), str(output_path / nifti_slice.name))
            break
    if delete_eyeballs:
        shutil.rmtree(str(PATHS['slices']), ignore_errors=True)
        shutil.rmtree(str(PATHS['eyeballs']), ignore_errors=True)

def temporalis_csa(input_path: Path = PATHS['temporalis'],
                    output_path: Path = PATHS['main'] / 'temporalis_csa.csv',
                    *, both_sides: bool, delete_all: bool = False) -> None:
    """
    Calculate the temporalis csa and output it in a csv

    Binarize the temporalis segmentation contained in input_path/patient/segmenttion,
    and count the white pixel (corresponding to the temporalis CSA).
    Input all the information in a csv of composed of the following coloumns:
        - patients_name: extracted from the directory name
        - slices_nb: extracted from the filename minus the extension
        - *_areas_temporalis: extracted from this function
    Output it in a csv file with the desired name

    Args:
        input_path (Path): Root path where the segmentation of all patients are stored
        output_path (Path): Path to where the csv will be saved
        csv_name (str): String of the desired output filename
    Kwargs:
        both_sides (bool): If you want to input both temporalis CSA
                            (need to segment both sides beforehand)
        delete (bool): Delete all the files except the csv
    """
    patients_name, slices_nb, l_areas_temporalis, r_areas_temporalis = ([], [], [], [])
    for patient in input_path.iterdir():
        patients_name.append(patient.name)
        segmentation_path = patient / 'segmentation'
        segmentations = list(segmentation_path.iterdir())
        nb_segmentations = len(segmentations)
        # both_sides = True if nb_segmentations == 2 else False
        if nb_segmentations == 3:
            # this case happen if you run once with both sides and after with
            # just one side or the other way around
            # the end result will be nb.png, l_nb.png, r_nb.png present in the fodler
            # we will use the fact that a number will always be before a letter
            # if we refer to the ASCII table
            if both_sides:
                segmentations[0].unlink()
            else:
                segmentations[1].unlink()
                segmentations[2].unlink()
        if both_sides:
            slices_nb.append(int(segmentations[1].stem[len('l_'):]))
        else:
            slices_nb.append(int(segmentations[0].stem))
        for segmentation in segmentation_path.iterdir():
            img = cv2.imread(str(segmentation))
            gray = cv2.cvtColor(img,cv2.COLOR_BGR2GRAY)
            _, thresh = cv2.threshold(gray, 127, 255, cv2.THRESH_BINARY)
            if both_sides:
                if segmentation.stem.startswith('l_'):
                    l_areas_temporalis.append(cv2.countNonZero(thresh))
                elif segmentation.stem.startswith('r_'):
                    r_areas_temporalis.append(cv2.countNonZero(thresh))
            else:
                l_areas_temporalis.append(cv2.countNonZero(thresh))
                r_areas_temporalis.append('')

    if delete_all:
        shutil.rmtree(str(PATHS['slices']), ignore_errors=True)
        shutil.rmtree(str(PATHS['eyeballs']), ignore_errors=True)
        shutil.rmtree(str(PATHS['temporalis']), ignore_errors=True)
    temporalis = {
        'Patient': patients_name,
        'Slice_nb': slices_nb,
        'Temporalis_CSA_left (in px)': l_areas_temporalis,
        'Temporalis_CSA_right (in px)': r_areas_temporalis
    }
    # if both_sides:
    #     temporalis['Temporalis_CSA_right (in px)'] = r_areas_temporalis
    temporalis_df = pd.DataFrame(temporalis)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    temporalis_df.to_csv(output_path)
    log.info(f"\nCreated {output_path.name}")



# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.


# /!\ only for specific use, not intended for the default usage of the package

# make segmenting new images easier
# 0) run prequisites if necessary
# 1) run manual_pipeline()
# 2) run prepare_segmentation(dataset="name_of_dataset")
# 3) make segmentation with itksnap and save them with same name but append _seg
# 4) run convert_segmentation()

from pathlib import Path

import shutil
import numpy as np
import pandas as pd
import imageio
import nibabel as nib

from tcsa import PATHS
from .prequisite import run_prequisite
from .pipeline import manual_pipeline, select_slice_patients


def prepare_segmentation(input_path: Path = PATHS['temporalis'],
                         output_path: Path = PATHS['main'] / 'segmentation',
                         *, dataset: str = 'Unknown') -> None:
    """
    Take input from the middle of the pipeline (just after the slice selection),
    proceed to copy the correct slice in png and nifti format in a new directory
    called 'segmentation' (output_path) and renaming it as an integer.
    Keep a record of the original patient name (patient_names),
    and the slice number (slices_nb), and the new integer name (file_nb)
    and proceed to save it in a csv file with the name of the dataset

    Args:
        input_path (Path): Path to were the right slice is stored
        output_path (Path): Path to were the images used for manual segmentation
                            will be stored
    Kwargs:
        dataset (str): Name of the dataset, that will be used for the csv filename
    """
    patient_names = []
    slices_nb = []
    file_nb = []
    output_path_png = output_path / 'png'
    output_path_nii = output_path / 'nifti'
    output_path_png.mkdir(parents=True, exist_ok=True)
    output_path_nii.mkdir(parents=True, exist_ok=True)
    for i, patient in enumerate(input_path.iterdir()):
        patient_names.append(patient.name)
        png_path = list((patient / 'original').iterdir())[0]
        slices_nb.append(int(png_path.stem))
        file_nb.append(i)
        nifti_path = PATHS['slices'] / patient.name / 'preprocessed' / f'{png_path.stem}.nii.gz'
        shutil.copy(str(png_path), str(output_path_png / f'{i}.png'))
        shutil.copy(str(nifti_path), str(output_path_nii / f'{i}.nii.gz'))
    segmentation = {
        'Patient': patient_names,
        'Slice_nb': slices_nb,
        'Filename': file_nb
    }
    segmentation_df = pd.DataFrame(segmentation)
    segmentation_df.to_csv(output_path / f'{dataset}.csv')

def convert_segmentation(input_path: Path = PATHS['main'] / 'segmentation') -> None:
    """
    Convert manualy segmented nifti slice into a png file

    Args:
        input_path (Path): root directory where the segmentation are stored
    """
    for segmentation in (input_path / 'nifti').glob('*_seg.nii.gz'):
        img = nib.load(segmentation).get_fdata()
        img = np.flipud(np.rot90(img[:,:]))
        filename = segmentation.name[: -len(''.join(segmentation.suffixes))]
        output_path = input_path / 'png' / f'{filename}.png'
        imageio.imsave(str(output_path), img)

# run_prequisite()
# manual_pipeline()
# prepare_segmentation(dataset='UKBB')
# convert_segmentation()



# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.


import argparse
from pathlib import Path
import shutil
import sys

from tcsa import welcome, __version__
from .prequisite import create_parser, handle_dicom, handle_single_nifti
from .pipeline import full_pipeline

parser = create_parser()

if len(sys.argv)==1:
    parser.print_usage()
    sys.exit(0)

args = parser.parse_args()

if args.version:
    welcome()
    sys.exit(0)

args.images = args.images.resolve(strict=True) # check if the path exist
args.output = args.output.resolve(strict=False)

if args.dicom:
    handle_dicom(args, verbosity=False)

if args.images.is_file() and not args.dicom:
    handle_single_nifti(args)

def main() -> None:
    """
    Run the full script.
    """
    # welcome()
    full_pipeline(input_path=args.images, output_path=args.output, delete=args.delete, both_sides=args.both_sides)
    shutil.rmtree(str(Path('temp_images').resolve()), ignore_errors=True)



# Copyright (C) 2022 Imperial College London
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/.


from pathlib import Path

from tcsa import PATHS
from .prequisite import run_prequisite
from .data_preparation import (slicing, preprocessing, converting_to_png)

from .model import unet
from .data import (testGenerator, saveResult)
from .thresholding import (select_slice, temporalis_csa)
import cv2

import logging
FORMAT = "%(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
log = logging.getLogger(__name__)

def segment_eyeballs() -> None:
    """
    Segment the eyeballs from the png slices present in eyeballs/patient/original,
    and save the result in eyeballs/patient/segmentation as png.
    """
    model = unet(PATHS['weights'] / 'unet_eyeball.hdf5')
    log.info("\nSegmenting eyeballs ...")
    for patient in (PATHS['eyeballs']).iterdir():
        slices_path = patient / 'original'
        testGene = testGenerator(slices_path)
        number_of_test_images = len(list(slices_path.iterdir()))
        names = [file.name for file in slices_path.iterdir()]
        log.info(f'Segmenting eyeball for {patient.name}...')
        results = model.predict_generator(testGene,number_of_test_images,verbose=1)
        saveResult(patient / 'segmentation', results, names)

def select_slice_patients(*, delete_eyeballs: bool = False) -> None:
    """
    Iterate throught all the patients, and run select_slice for each patient.

    Kwargs:
        delete_eyeballs (bool): pass to select_slice,
                                delete all slices (eyeball png and nifti slices)
    """
    for patient in (PATHS['eyeballs']).iterdir():
        segmentation_path = patient / 'segmentation'
        select_slice(segmentation_path,
            PATHS['temporalis'] / patient.name, delete_eyeballs=delete_eyeballs)

def segment_temporalis(*, both_sides: bool = False) -> None:
    """
    Segment the temporalis from the png slice present in temporalis/patient/original
    and save the result in temporalis/patient/segmentation as png

    Kwargs:
        both_sides (bool): if set to true will segment both sides
                           (the other sides will be flipped)
    """
    model = unet(PATHS['weights'] / 'unet_temporalis.hdf5')
    log.info("\nSegmenting temporalis ...")
    for patient in PATHS['temporalis'].iterdir():
        slice_path = patient / 'original'
        if both_sides: right_side(slice_path)
        testGene = testGenerator(slice_path)
        number_of_test_images = 1 if not both_sides else 2
        names = [file.name for file in slice_path.iterdir()]
        log.info(f'Segmenting temporalis for {patient.name}...')
        results = model.predict_generator(testGene,number_of_test_images,verbose=1)
        saveResult(patient / 'segmentation', results, names)

def right_side(input_path: Path) -> None:
    """
    flip the image and create a copy to be able to segment both sides.
    the original one will be renamned l_nbslice.png and the flipped r_nbslice.png

    Args:
        input_path (Path): path provided by segment_temporalis, it's the path
                           to temporalis/patient/original
    """
    for slice in input_path.iterdir():
        if slice.name.startswith(('l_', 'r_')): return
        img = cv2.imread(str(slice))
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        gray_flipped = cv2.flip(gray, 1)
        cv2.imwrite(str(slice.parent / f'r_{slice.name}'), gray_flipped)
        slice.rename(slice.parent / f'l_{slice.name}')


def full_pipeline(input_path: Path, output_path: Path,
                  delete: bool = False, both_sides: bool = True) -> None:
    """
    Run the full pipeline.

    1) Prequisite: creates directories, and check weights for model
    2) Slicing: get the slices from the mri
    3) Preprocess: preprocess the slices
    4) Convert to PNG: needed for the neural network
    5) Segment eyeballs
    6) Select slice: select the slice at the eye level
    7) Segment temporalis: can segment both sides if needed
    8) Output: create a csv file with the patient name, the slice, and the CSA

    Args:
        input_path (Path): Path to the directory where the images are stored
        output_path (Path): Path to where the csv file will be saved,
                            (can be the directory with or without the filename)
        delete (bool): Delete all intermediate directories and files,
                       only keep the csv, images in input_path and weights
        both_sides (bool): Segment both sides, only left side by default
    """
    run_prequisite()

    nb_patients = len(list(input_path.iterdir()))
    log.info(f"\nYou have {nb_patients} image{'s' if nb_patients > 1 else ''}")

    slicing(input_path)
    preprocessing(delete_og_slices=False)
    converting_to_png()

    segment_eyeballs()
    select_slice_patients(delete_eyeballs=False)
    segment_temporalis(both_sides=both_sides)

    temporalis_csa(output_path=output_path, both_sides=both_sides, delete_all=delete)

def manual_pipeline() -> None:
    """
    Run only the necessary part of the pipeline to make manual segmentation
    /!\ Only for specific use
    """
    slicing()
    preprocessing(delete_og_slices=False)
    converting_to_png()

    segment_eyeballs()
    select_slice_patients(delete_eyeballs=False)
